/* -*- mode: c++; c-basic-offset: 3; indent-tabs-mode: nil; -*- */
// System Header Files   
#include <cstdio>

#include <algorithm>
#include <stdexcept>   
#include <memory>   
#include <iterator>   
#include <limits>
#include <sstream>
   
#include "general/Thread.hh"

// Local Header Files
#include "diskcacheAPI.hh"
#include "dirhash.hh"
#include "dirinfo.hh"      
#include "cachefile.hh"   
#include "HotDirectory.hh"
#include "mountpt.hh"   
#include "MountPointScanner.hh"

using namespace diskCache;
using namespace General;   
using namespace std;

   
static General::ReadWriteLock::lock_type exit_lock =
   General::ReadWriteLock::Sync::INITIALIZER;

#if 0
#include "general/IOLock.hh"

#define VERBOSE_DEBUGGING 1
#define	AT_BASE( a ) \
{ \
   int oldstate; \
   pthread_setcancelstate( PTHREAD_CANCEL_DISABLE, &oldstate ); \
   MutexLock    lock( General::IOLock::GetKey( std::cerr ) ); \
   pthread_setcancelstate( oldstate, &oldstate ); \
\
   std::cerr << a << __FILE__ << " " << __LINE__ << std::endl << std::flush; \
}

#else
#define	AT_BASE( a )
#endif

#define WORKING_WRITE_QUEUE_BLOCKING 0

#define	AT( a )	AT_BASE( a << std::endl << "\t" )
#define	HERE( ) AT_BASE( "" )

namespace
{
   static const CHAR Slash( '/' );   
}

   
MutexLock::lock_type DirectoryHash::m_scanned_directories_baton
      = MutexLock::Initialize( );

INT_4U DirectoryHash::m_scanned_directories = 0;

MutexLock::lock_type DirectoryHash::m_write_queue_baton
      = MutexLock::Initialize( );

const INT_4U DirectoryHash::SCANNED_DIRECTORIES_RESET
= std::numeric_limits< INT_4U >::max( );

//: This macro will instantiate static data members and methods for
//: a "singleton" part of DirectoryHash class.
DEFINE_READ_WRITE_SINGLETON( DirectoryHash, 75 )
   
   
//------------------------------------------------------------------------------
//   
//: Constructor
//
//!param: const list< string >& dirs - Known directory names to be excluded.
//
diskCache::DirectoryHash::IsExcluded::IsExcluded( 
   const dir_container_type& dirs )
   : mDirNames( dirs )
{}
   
   
//------------------------------------------------------------------------------
//      
//: Compare operator
//
//!param: const string& path - Full path of the directory to examine.
//
//!return: bool - True if directory name is found in mDirNames list,
//+        false otherwise.
//
bool diskCache::DirectoryHash::IsExcluded::operator()( const std::string& path )
   const
{
   // Extract directory name from full path
   std::string::size_type name_pos = path.find_last_of( Slash );
   
   // if path doesn't contain any slashes (should never happen)
   // OR path = "/"
   if( name_pos == std::string::npos || path.size() == 1 )
   {
      return false;
   }

      
   const string name( path.substr( ++name_pos ) );
 
   return ( find( mDirNames.begin(), mDirNames.end(), name ) != mDirNames.end() );
}


//------------------------------------------------------------------------------
//   
//: Constructor
//
//!param: const string& parent - Parent directory name.
//
diskCache::DirectoryHash::HasParent::
HasParent( const std::string& parent )
   : mParent( parent )
{
   mParent += Slash;
   mParentSize = mParent.size();
}


//------------------------------------------------------------------------------
//   
//: Destructor
//
diskCache::DirectoryHash::HasParent::~HasParent()
{}
   
  
//------------------------------------------------------------------------------
//   
//: Compare operator
//
//!param: const string& path - Full path of the directory to examine.
//
//!return: bool - True if directory has specified parent directory,
//+        false otherwise.
//
bool diskCache::DirectoryHash::HasParent::
operator()( const std::string& path ) const
{
   // "compare" takes care if mParentSize > (path.size - 0)
   return ( mParent.compare( 0, mParentSize, path, 0, mParentSize ) == 0 );
}   
   

const INT_4S diskCache::DirectoryHash::
GetMountPointMTime( const std::string& MountPoint,
                    const dir_container_type& ExcludeDirs )
{
#if VERBOSE_DEBUGGING
   static const char method_name[] = "diskCache::DirectoryHash::GetMountPointMTime";
#endif /* VERBOSE_DEBUGGING */
   General::ReadWriteLock       elock( exit_lock,
                                       General::ReadWriteLock::READ );
   LogMessage errors;

   AT( method_name << ": Entry" );

   // To handle MOUNT_PT entry matching orwne of "exclude_dirs"
   // Check for directories that are pending for deletion: do it once at
   // the beginning of MOUNT_PT entry scan.
   list< string > pending_dirs;
   {
      ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
                             __FILE__, __LINE__ );   
   }
   
   for( list< string >::const_iterator pending_iter = pending_dirs.begin(),
           pending_end_iter = pending_dirs.end();
        pending_iter != pending_end_iter; ++pending_iter )
   {
      resulting_dirs_container_type        rm_dirs;

      removeDirectory( *pending_iter,
                       rm_dirs,
                       errors );
   }
   
   
   IsExcluded excluded( ExcludeDirs );
   
   if ( excluded( MountPoint ) )
   {
      return -1;
   }
   
   
   ReadWriteLock lock( mLock, ReadWriteLock::READ,
                       __FILE__, __LINE__ );   
   
   const_iterator iter( mAllDirectories.find( MountPoint ) );

   AT( method_name << ": obtained iter for: " << MountPoint );
   // There is no such directory entry, insert new one.
   if( iter == mAllDirectories.end() )
   {
      return -1;
   }
   try
   {
      return iter->second->TimeModified( );
   }
   catch( ... )
   {
      return -1;
   }
}

void diskCache::DirectoryHash::
AddDirectory( General::SharedPtr< DirectoryInfo > DirInfo )
{
   SharedPtr< DirectoryInfo >   old;

   //----------------------------------------------------------------------------
   // Modify local information
   //----------------------------------------------------------------------------
   {
      //-------------------------------------------------------------------------
      // Look up the directory information
      //-------------------------------------------------------------------------
      ReadWriteLock     lock( mLock, ReadWriteLock::WRITE,
                              __FILE__, __LINE__ );
   
      
      iterator iter( mAllDirectories.find( DirInfo->GetPath( ).Var( ) ) );

      if ( iter != mAllDirectories.end( ) )
      {
         //----------------------------------------------------------------------
         // There is an entry.
         // 1) Get a reference to the old data
         // 2) Replace with the new data
         //----------------------------------------------------------------------
         old = iter->second;
         iter->second = DirInfo;
      }
      else
      {
         //----------------------------------------------------------------------
         // No entry exists
         // 1) Add new data
         //----------------------------------------------------------------------
         mAllDirectories[ DirInfo->GetPath( ).Var( ) ] = DirInfo;
      }
   }
   //----------------------------------------------------------------------------
   // Do this as an atomic operation.
   //----------------------------------------------------------------------------
   diskCache::MountPointHash::Replace( old, DirInfo );
}

void diskCache::DirectoryHash::
RemoveDirectory( const std::string& DirectoryName )
{
   SharedPtr< DirectoryInfo >   old;

   {
      //-------------------------------------------------------------------------
      // Look up the directory information
      //-------------------------------------------------------------------------
      ReadWriteLock     lock( mLock, ReadWriteLock::WRITE,
                              __FILE__, __LINE__ );   
      
      iterator iter( mAllDirectories.find( DirectoryName ) );

      if ( iter == mAllDirectories.end( ) )
      {
         //----------------------------------------------------------------------
         // Nothing to be removed
         //----------------------------------------------------------------------
         return;
      }
      //-------------------------------------------------------------------------
      // There is an entry.
      //----------------------------------------------------------------------
      old = iter->second;
      mAllDirectories.erase( iter );
   }
   if ( old )
   {
      //-------------------------------------------------------------------------
      // Remove from the mount point hash
      //-------------------------------------------------------------------------
      diskCache::MountPointHash::Unregister( old );
   }
}

//------------------------------------------------------------------------------
//
//: Add/rescan directory record in the global hash map.
//   
// This method looks up a directory path in the existing global hash map.
// It inserts new record if the hash doesn't have an entry for it, and collects
// necessary directory metadata for further lookups.   
//   
// Method recursively rescans directory for new/deleted frame data for
// already existing unordered_map entry.   
//
//!param: const std::string& dir - Full path for the directory.
//!param: const std::string& mount_pt - Full path for the mount point directory 
//+       belongs to.
//!param: const std::list< std::string >& exclude_dirs - A list of known 
//+       directory names that should not be scanned.
//!param: const std::string& errors - Accumulative error message.   
//!param: const INT_4S new_mtime - Mtime for the "dir" at the time method is 
//+       called. Default is -1: unknown mtime.
//   
//!return: std::list< std::string > - List of scanned subdirectories
//+        that contain frame data.
//   
const std::list< std::string > diskCache::DirectoryHash::
scanDirectory( const std::string& Directory,
               const std::string& MountPoint,
               const dir_container_type& ExcludeDirs,
               LogMessage& errors,
               const INT_4S new_mtime )   
{
   static const CHAR* const method_name( "DirectoryHash::scanDirectory(): " );

   std::string  reason_for_cancellation;

   //--------------------------------------------------------------------
   // Check to see if this scan should terminate immediately
   //--------------------------------------------------------------------
   if ( MountPointScanner::CancellationRequest( reason_for_cancellation ) )
   {
      //-----------------------------------------------------------------
      // Setup for cancellation but don't actually do the cancellation
      //-----------------------------------------------------------------
      General::Thread::Self::Cancel( );
      //-----------------------------------------------------------------
      // Log reason for the cancellation
      //-----------------------------------------------------------------
      GenericAPI::queueLogEntry( reason_for_cancellation,
                                 GenericAPI::LogHTML::MT_DEBUG,
                                 30,
                                 method_name,
                                 "SCAN_MOUNTPT" );
   }
   //--------------------------------------------------------------------
   // Give the thread a chance to terminate cleanly.
   //--------------------------------------------------------------------
   General::Thread::Self::CancellationCheck( method_name,
                                             __FILE__, __LINE__ );
   
   //--------------------------------------------------------------------
   // Start the real work of scanning the directory.
   //--------------------------------------------------------------------
   General::ReadWriteLock       elock( exit_lock,
                                       General::ReadWriteLock::READ );
   
   AT( method_name << ": Entry" );

   // To handle MOUNT_PT entry matching one of "exclude_dirs"
   if( Directory == MountPoint )
   {
      IsExcluded excluded( ExcludeDirs );
   
      if( excluded( Directory ) )
      {
         AT( "scanDirectory: excluding dir: " << Directory );
         const list< string >
            removed_subdirs( MountPointHash::removeEntry( Directory, errors ) );
   
         if( removed_subdirs.empty() == false ) 
         {
            // Log list of deleted subdirectories as a result of 
            // directory removal:   
            // FIX for PR 2603: opening brace needs preceding space
            errors( )
               << " {";
      
            errors( ) << MountPointHash::createExcludedMsgHeader( ExcludeDirs );
      
            copy( removed_subdirs.begin(), removed_subdirs.end(),
                  ostream_iterator< string >( errors( ), " " ) );      
      
            errors( ) << "} ";   
         }   
   
         return list< string >();
      }
   }
   
   std::list< std::string >  retval;
   std::list< std::string >  dir_queue;

   dir_queue.push_front( Directory );
   while( ( dir_queue.empty( ) == false )
          && ( General::AtExit::IsExiting( ) == false ) )
   {
      std::string       dir( dir_queue.front( ) );

      dir_queue.pop_front( );

      ScannedDirectories( 1 );

      // It's kind of dangerous to expose this pointer outside of lock 
      // if MOUNT_PT list gets trimmed and Tcl starts removing 
      // directories while there are still some active threads that are
      // using those directories.
      // If Tcl can't guarantee that it'll wait for those active threads to
      // finish before creating removal threads ---> need to do all 
      // DirectoryInfo manipulation under the lock of DirectoryHash 
      // (lock within DirectoryInfo is not enough)!
      //
      // Update: Aug 14, 2009
      //   This is now safe with the use of shared pointers.
      mapped_type info;

      bool is_registered_hot = HotDirectory::IsRegistered( dir );

      AT( method_name << ": About to do descending scan for dir: " << dir );
      // Introduce this scope for WRITE lock
      {
#if WORKING_WRITE_QUEUE_BLOCKING
         MutexLock         wlock( m_write_queue_baton, __FILE__, __LINE__ );
#endif /* WORKING_WRITE_QUEUE_BLOCKING */
         ReadWriteLock     lock( mLock, ReadWriteLock::READ,
                                 __FILE__, __LINE__ );   
   
         const_iterator iter( mAllDirectories.find( dir ) );      

         AT( method_name << ": obtained iter for: " << dir );
         // There is no such directory entry, insert new one.
         if( iter == mAllDirectories.end() )
         {
            AT( method_name << ": iterator is at the end" );

            //----------------------------------------------------------------------
            // Change to write lock since the list most likely will be modified.
            //----------------------------------------------------------------------
            lock.Modify( ReadWriteLock::WRITE, __FILE__, __LINE__ );
   
            iter = mAllDirectories.find( dir );

            // There is no such directory entry, insert new one.
            if( iter == mAllDirectories.end( ) )
            {
               // Insert new directory entry
               General::SharedPtr< DirectoryInfo >
                  di( new DirectoryInfo( dir, MountPoint ) );
               pair< iterator, bool >
                  insert_result( add_directory( di ) );
   
               // Get pointer to the dir info object:
               info = insert_result.first->second;
            }
            else
            {
               // Another thread has put it into the list.
               info = iter->second;
            }
         }
         else
         {
            // Hash already has an entry for the directory
            AT( "Hash already has entry for: " << dir );
            info = iter->second;
         }
      }
      bool modified = false;
      std::list< std::string > scan_msgs( info->scan( ExcludeDirs, errors, modified ) );
      retval.splice( retval.end( ), scan_msgs );
      //-------------------------------------------------------------------------
      // Push to the front of the list the next set of directories to scan
      //-------------------------------------------------------------------------
      if ( modified )
      {
         {
            ReadWriteLock     lock( mLock, ReadWriteLock::READ,
                                    __FILE__, __LINE__ );   
   
            const_iterator iter( mAllDirectories.find( dir ) );
            if ( iter != mAllDirectories.end( ) )
            {
               info = iter->second;
            }
            else
            {
               info.reset( );
            }
         }
         //--------------------------------------------------------------
         // Signal that the information needs to be written to disk
         //--------------------------------------------------------------
         MountPointHash::LastUpdateSet( );
      }
      if ( info
           && ( info.unique( ) == false ) )
      {
         HotDirectory::RegisterIfHot( info, ExcludeDirs, !is_registered_hot );
         info->GetSubDirectories( dir_queue, dir_queue.begin( ) );
      }
   }

   
   // Scan directory for new/updated metadata:
   AT( method_name << ": Exit" );
   return retval;
}

   
//------------------------------------------------------------------------------
//
//: Remove directory from global hash.
//   
// This method removes directory and all its recursive subdirectories from
// global hash.
//
//!param: const std::string& dir - Full path for the directory to remove.
//!param: std::string& errors - Accumulative error message.   
//!param: const bool& unregister - A flag to indicate if directory needs to 
//+       unregister itself with global frame hash. Default is FALSE (set to
//+       FALSE only when MOUNT_PT directory is removed ===> the whole entry 
//+       for MOUNT_PT directory is removed from frame hash at once).   
//   
//!return: const std::list< std::string > - List of removed
//+        subdirectories with frame data. This list will be empty if no data
//+        was removed as a result of directory removal.
// 
//!exc: runtime_error - Directory doesn't exist.
//   
int diskCache::DirectoryHash::
removeDirectory( const std::string& dir,
                 resulting_dirs_container_type& RemovedDirectories,
                 LogMessage& errors,
                 const bool& unregister )
{
   mapped_type info;
   
   // Introduce this scope for WRITE lock
   {
#if WORKING_WRITE_QUEUE_BLOCKING
      MutexLock         wlock( m_write_queue_baton, __FILE__, __LINE__ );
#endif /* WORKING_WRITE_QUEUE_BLOCKING */
      ReadWriteLock     lock( mLock, ReadWriteLock::WRITE,
                          __FILE__, __LINE__ );      
   
      iterator iter( mAllDirectories.find( dir ) );
      if( iter == mAllDirectories.end() )
      {
         // Tcl layer says be silent about non-existing directories in the hash
#ifdef REPORT_MISSING_DIR   
         static const CHAR* const method_name( "DirectoryHash::removeDirectory(): " );   
         // FIX for PR 2603: opening brace needs preceding space
         errors( LogMessage::LM_NOTE ) << " {"
                   << method_name
                   << " directory \'"
                   << dir
                   << "\' doesn't exist in global hash.} ";
#endif   
   
         return 0;
      }
   
      // Keep directory object for automatic cleanup 
      info = iter->second;
   
      // Remove this DirectoryInfo object from DirectoryHash map
      mAllDirectories.erase( iter );
   }
   
   // FIX for PR 2455 and PR 2526:
   // need to use accumulative parameter 'errors' to capture sub dir removal 
   // messages since generation of 'REMOVE' message has been relocated to 
   // 'removeSubDirectories' function.

   // Remove all recursive subdirectories from global hash,
   // report removed subdirectories with frame data only.

   info->removeSubDirectories( RemovedDirectories, errors, unregister );
   if ( unregister )
   {
      MountPointHash::Unregister( info );
   }
   return info->FileCount( );
}


diskCache::DirectoryHash::
~DirectoryHash( )
{
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
diskCache::DirectoryHash::resulting_dirs_container_type diskCache::DirectoryHash::
ExcludeDirectories( const excluded_dirs_container_type& AddedToList,
                    const excluded_dirs_container_type& RemovedFromList,
                    LogMessage& errors )  
{
   typedef resulting_dirs_container_type return_type;
   return_type all_removed_subdirs;

   //--------------------------------------------------------------------
   // Check to see if any work needs to be done
   //--------------------------------------------------------------------
   if( ( AddedToList.empty( ) == true )
       && ( RemovedFromList.empty( ) == true ) )
   {
      //-----------------------------------------------------------------
      // Since there isn't, just return the empty list.
      //-----------------------------------------------------------------
      return all_removed_subdirs;
   }
   
   
   //--------------------------------------------------------------------
   // Get list of mount points
   //--------------------------------------------------------------------
   std::list< dir_data_type >     mount_points;
   {
      ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
                          __FILE__, __LINE__ );            
      
      for( iterator
              iter = mAllDirectories.begin(),
              end_iter = mAllDirectories.end();
           iter != end_iter;
           ++iter )      
      {
         if ( iter->second->isMountPtEntry( ) )
         {
            mount_points.push_back( iter->second );
         }
      }
   }
   IsExcluded   exclude_check( AddedToList );

   while( mount_points.empty( ) == false )
   {
      dir_data_type     e( mount_points.front( ) );
      mount_points.pop_front( );
      if ( exclude_check( e->GetPath( ).Var( ) ) == true )
      {
         //--------------------------------------------------------------
         // Remove the entire mount point
         //--------------------------------------------------------------
         const return_type r( MountPointHash::removeEntry( e->GetPath( ).Var( ),
                                                           errors ) );

         all_removed_subdirs.insert( all_removed_subdirs.end( ),
                                     r.begin( ), r.end( ) );
      }
      else
      {
         std::string                    path;
         std::list< std::string >       paths;
         while( true )
         {
            //-----------------------------------------------------------
            // Check entries under this mount point for any that may be
            // affected by the change to the exclude list
            //-----------------------------------------------------------
            if ( e )
            {
               e->ExcludedDirs( AddedToList,
                                RemovedFromList,
                                all_removed_subdirs,
                                errors );
               //--------------------------------------------------------
               // Get any subdirectories and put them onto the stack to
               // be processed
               //--------------------------------------------------------
               e->GetSubDirectories( paths, paths.begin( ) );
               //--------------------------------------------------------
               // Done with the resource, so dispose of it.
               //--------------------------------------------------------
               e.reset( );
            }
            //-----------------------------------------------------------
            // Check to see if there is more work to be done
            //-----------------------------------------------------------
            if ( paths.empty( ) )
            {
               break;
            }
            //-----------------------------------------------------------
            // Grab the head of the list and find the node associated
            // with it.
            //-----------------------------------------------------------
            {
               ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
                                   __FILE__, __LINE__ );            
               do
               {
                  path = paths.front( );
                  paths.pop_front( );
                  container_type::const_iterator pos( mAllDirectories.find( path ) );

                  if ( pos != mAllDirectories.end( ) )
                  {
                     e = pos->second;
                  }
               } while( ( ! e )
                        && ( paths.empty( ) == false ) );
            }
         } 
      }
   }

   return all_removed_subdirs;
}

   
//------------------------------------------------------------------------------
//
//: Exclude directories from global hash.
//   
// This method removes directories that match specified excluded directory list
// and all its recursive subdirectories from global hash. If excluded directory
// is a subdirectory, it moves that excluded subdirectory from
// DirectoryInfo::mSubDirs into DirectoryInfo::mExcludedSubDirs. If excluded
// directory is one of the MOUNT_PT entries, the whole entry will be removed
// from MountPointHash::mMountPtHash.   
//
//!param: const std::string& dir - Full path for the directory to exclude.
//!param: std::string& errors - Accumulative error message.   
//   
//!return: const std::list< std::string > - List of removed
//+        subdirectories with frame data. This list will be empty if no data
//+        was removed as a result of directory removal.
//   
const diskCache::DirectoryHash::resulting_dirs_container_type diskCache::DirectoryHash::
excludeDirectories( const excluded_dirs_container_type& dirs,
                    LogMessage& errors )  
{
   typedef resulting_dirs_container_type return_type;

   // There are no directories to exclude
   if( dirs.empty() == true )
   {
      return return_type();
   }
   
   
   // Collect all registered directories in the map
   
   // List of known directory names will be sorted 
   // (all entries in the map are already sorted by the key (by default))
   return_type all_dirs;
   {
      ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
                          __FILE__, __LINE__ );            
      
      for( iterator
              iter = mAllDirectories.begin(),
              end_iter = mAllDirectories.end();
           iter != end_iter;
           ++iter )      
      {
         all_dirs.push_back( iter->first );
      }
   }


   return_type::iterator iter( all_dirs.begin() );
   return_type all_removed_subdirs;

   // Step through all directories in the map      
   while( ( iter = find_if( iter, all_dirs.end(), IsExcluded( dirs ) ) )
          != all_dirs.end() )      
   {
      // Found directory to exclude
      return_type just_removed( exclude_directory( *iter, errors ) );
   

      // Move all elements of "just_removed" list into result list 
      all_removed_subdirs.splice( all_removed_subdirs.end(), just_removed );
      
   
      const string parent_path( *iter );
   
      // Skip excluded directory (++iter) and removed subdirectories
      // from "all_dirs" list
   
      // This step relies on "all_dirs" list to be lexicographically sorted:
      // all subdirectories will appear right after parent directory
      iter = find_if( ++iter, all_dirs.end(), not1( HasParent( parent_path ) ) );            
   }
   
   
   return all_removed_subdirs;
}

   
//------------------------------------------------------------------------------
//
//: Remove all registered directories from global hash.   
//
// This method is called only at a time when new frame hash is about to be read
// from the file.
//   
//!return: Nothing.
//   
void diskCache::DirectoryHash::removeDirectories()
{
#if WORKING_WRITE_QUEUE_BLOCKING
   MutexLock            wlock( m_write_queue_baton, __FILE__, __LINE__ );
#endif /* WORKING_WRITE_QUEUE_BLOCKING */
   ReadWriteLock        lock( mLock, ReadWriteLock::WRITE,
                       __FILE__, __LINE__ );      
   
   mAllDirectories.clear();
   
   return;
}

///----------------------------------------------------------------------
/// \brief Write specific directory to the stream.
///   
/// This method writes a specific directory to the cache file recursively.
///   
/// \param dirname Directory to write to the file.
/// \param stream Stream to write to.   
///
/// \return Nothing.
///----------------------------------------------------------------------
void diskCache::DirectoryHash::
writeDirectory( const std::string& dirname, 
                CacheFile& stream )
{
   // Remember subdirectories
   list< string > subdirs;
   
   
   // Introduce this scope for READ lock
   {
      ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
                          __FILE__, __LINE__ );      
   
      iterator iter( mAllDirectories.find( dirname ) );
      if( iter == mAllDirectories.end() )
      {
         // Directory already doesn't exist
         return;
      }

      if( iter->second )
      {
         //--------------------------------------------------------------
         // Be cooperative with cancelation process
         //--------------------------------------------------------------
         Thread::Self::CancellationCheck( "diskCache::DirectoryHash::writeDirectory",
                                          __FILE__, __LINE__ );
         // A valid pointer
         iter->second->write( stream );
   
         subdirs = iter->second->getSubDirectoryList();
      }
   }   

   
   // Write all subdirectories
   typedef list< string >::const_iterator LSConstIter;
   
   for( LSConstIter iter = subdirs.begin(), end_iter = subdirs.end();
        iter != end_iter; ++iter )
   {
      writeDirectory( *iter, stream );
   }
   
   
   return;
}

   
///----------------------------------------------------------------------
/// \brief Write specific directory to the stream.
///   
/// This method writes a specific directory to the cache file recursively.
///   
/// \param dirname Directory to write to the file.
/// \param stream Stream to write to.   
///
/// \return Nothing.
///----------------------------------------------------------------------
void diskCache::DirectoryHash::
WriteDirectories( std::list< std::string >& Dirs,
                  CacheFile& Stream )
{
   while ( Dirs.empty( ) == false )
   {
      // Get the first element.
      const std::string dirname( Dirs.front( ) );
      Dirs.pop_front( );

      mapped_type       node;
      
      //
      // Introduce this scope for READ lock
      {
         ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
                             __FILE__, __LINE__ );      
   
         iterator iter( mAllDirectories.find( dirname ) );
         if( ( iter != mAllDirectories.end( ) )
             && ( iter->second.get( ) ) )
         {
            //--------------------------------------------------------------
            // Be cooperative with cancelation process
            //--------------------------------------------------------------
            Thread::Self::CancellationCheck( "diskCache::DirectoryHash::writeDirectory",
                                             __FILE__, __LINE__ );
            node = iter->second;
         }
      }   

      if ( node )
      {
         // A valid pointer
         node->write( Stream );

         if ( ! node.unique( ) )
         {
            const std::list< std::string >
               subdirs( node->getSubDirectoryList( ) );
   
            // Add additional directories to list of directories to write.
            Dirs.insert( Dirs.begin( ),
                         subdirs.begin( ),
                         subdirs.end( ) );
         }
      }
   }
}

   
//------------------------------------------------------------------------------
//
//: Read all directory entries from the stream.
//   
// This method reads all directory entries from the cache file.
//   
//!param: CacheFile& stream - Stream to read from.
//
//!return: Nothing.
//   
void diskCache::DirectoryHash::readDirectories( CacheFile& stream )
{
   mapped_type new_dir;

   while( stream.peek() != EOF )
   {
      new_dir = DirectoryInfo::read( stream );

      try
      {
         // Insert new entry into the global directory map:
         {
#if WORKING_WRITE_QUEUE_BLOCKING
            MutexLock      wlock( m_write_queue_baton, __FILE__, __LINE__ );
#endif /* WORKING_WRITE_QUEUE_BLOCKING */
            ReadWriteLock  lock( mLock, ReadWriteLock::WRITE,
                                 __FILE__, __LINE__ );         
   
            add_directory( new_dir );
         }       

         //--------------------------------------------------------------
         // Register the new data with the search cache.
         // Do not check for conflicts since the data is coming from
         //   a file. Only non-conflicting data is saved to the file.
         //--------------------------------------------------------------
         if ( new_dir->HasData( ) )
         {
            diskCache::MountPointHash::Register( new_dir );
         }
      }
      catch( const std::exception& Exception )
      {
         if ( new_dir )
         {
            std::cerr << new_dir->GetPath( ).Var( );
         }
         std::cerr
            << std::endl
            ;
      }
      catch( ... )
      {
         // Ignore any registry errors since mount point conflicts
         // may occur.
         std::cerr << "WARNING: diskCache::DirectoryHash::readDirectories"
                   << " registerFrameData: exception"
                   << std::endl
            ;
      }
   }
   
   
   return;
}

   

std::pair< diskCache::DirectoryHash::iterator, bool > diskCache::DirectoryHash::
add_directory( mapped_type Info )
{
   const std::string path( Info->GetPath( ).Var( ) );

   iterator cur = mAllDirectories.find( path );
   std::pair< iterator, bool > retval;

   if ( cur == mAllDirectories.end( ) )
   {
      retval = mAllDirectories.insert( value_type( path, Info ) );
   }
   else
   {
      cur->second = Info;
      retval = std::pair< iterator, bool >( cur, true );
   }

   return retval;
}

//------------------------------------------------------------------------------
//
//: Exclude directory from global hash.
//   
// This method removes directory and all its recursive subdirectories from
// global hash. If directory is a subdirectory, change metadata of the parent
// directory to have it listed as mExcludedSubDirs.   
//
//!param: const std::string& dir - Full path for the directory to remove.   
//!param: std::string& errors - Accumulative error message.      
//   
//!return: const std::list< std::string > - List of removed
//+        subdirectories with frame data. This list will be empty if no data
//+        was removed as a result of directory removal.
//   
const diskCache::DirectoryHash::resulting_dirs_container_type diskCache::DirectoryHash::
exclude_directory( const std::string& dir,
                   LogMessage& errors )
{
   typedef resulting_dirs_container_type        return_type;

   // If excluded directory is a MOUNT_PT entry, remove the whole entry 
   // from MountPointHash without resetting metadata for its parent directory
   bool is_mount_pt_entry( false );
   
   {
      ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
                          __FILE__, __LINE__ );               
   
      iterator iter( mAllDirectories.find( dir ) );
      if( iter != mAllDirectories.end() )
      {
         is_mount_pt_entry = iter->second->isMountPtEntry();
      }
   }
   

   if( is_mount_pt_entry )
   {
      // Remove the whole mount point from frame hash
      return MountPointHash::removeEntry( dir, errors );
   }

   
   // Directory is a subdirectory: update mExcludedSubDirs of the parent
   // Extract parent name
   const string::size_type slash_pos( dir.find_last_of( Slash ) );
   const string parent( dir, 0, slash_pos );
   
   {
      ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
                          __FILE__, __LINE__ );               
   
      iterator iter( mAllDirectories.find( parent ) );
      if( iter != mAllDirectories.end() )
      {
         // If MOUNT_PT entry parent belongs to is being scanned --> risky to
         // to remove it's subdirectory, let deep scan to handle it next time around.
         const string parent_mount_pt( iter->second->GetMountPointPath( ).Var( ) );
   
         iterator mount_pt_iter( mAllDirectories.find( parent_mount_pt ) );
         if( mount_pt_iter == mAllDirectories.end() )
         {
            // FIX for PR 2603: opening brace needs preceding space
            errors( ) <<" {Missing MOUNT_PT entry for directory to exclude: \'"
                      << dir
                      << "\'} "
               ;
   
            return return_type();
         }
   
         iter->second->setExcludedSubDir( dir );
      }
   }
   
   
   // Delete directory recursively
   static const bool unregister( true );
   
   return_type   rm_dirs;

   removeDirectory( dir,
                    rm_dirs,
                    errors,
                    unregister );
   return rm_dirs;
}
