/* -*- mode: c++; c-basic-offset: 3; indent-tabs-mode: nil; -*- */

// System Header Files   
#include <sys/time.h>   
#include <time.h>   

#include <iostream>
#include <iomanip>
#include <list>

#include "diskcachecmd.hh"

#include "general/AtExit.hh"
#include "general/ConditionalVariable.hh"
#include "general/ldasexception.hh"
#include "general/Task.hh"
#include "general/TaskThread.hh"
#include "general/ThreadPool.hh"

#if FOR_TCL
// GenericAPI Header Files
#include <genericAPI/thread.hh>
#include <genericAPI/tid.hh>
#endif /* FOR_TCL */
#include "genericAPI/Logging.hh"

// Local Header Files
#include "DumpCacheDaemon.hh"
#include "mountpt.hh"
#include "MountPointScanner.hh"
#include "dirhash.hh"
#if old
#include "framequery.hh"
#endif /* old */
#include "cachefile.hh"
#include "Query.hh"
#include "QueryResultsDump.hh"
#include "QueryResultsTimeRange.hh"
#include "ScanMountPointsDaemon.hh"
#include "SDGTxFilenameHandler.hh"


using std::ios_base;
using std::istringstream;
using std::ostringstream;
using std::runtime_error;
using std::string;

using namespace diskCache;
using General::ifstream;
using GenericAPI::LogEntryGroup_type;
using GenericAPI::queueLogEntry;
   
   
namespace {
   typedef diskCache::MountPointScanner::mount_point_container_type
   mount_point_container_type;

   //--------------------------------------------------------------------
   /// \brief Debug utility to track down execution time.   
   //--------------------------------------------------------------------
   class Timer
   {
   public:
    
      Timer();
      REAL_4 getDelta( const std::string& method );
   
   private:
    
      struct timeval mTv;
      INT_4S mLastSec;
      INT_4S mLastUsec;   
   };
   

   Timer::Timer()
   {
      ::gettimeofday( &mTv, NULL );
      mLastSec = mTv.tv_sec;
      mLastUsec = mTv.tv_usec;      
   }
   
   
   REAL_4 Timer::
   getDelta( const std::string& method )
   {
      ::gettimeofday( &mTv, NULL );
   
      REAL_4 delta( mTv.tv_sec - mLastSec + ( mTv.tv_usec - mLastUsec ) * 1e-6 );
   
#ifdef DEBUG_TIME      
      cout << "--->DEBUG_TIME: clock time {" << method << "}: "
           << delta << endl;   
   
      mLastSec = mTv.tv_sec;
      mLastUsec = mTv.tv_usec;
#endif   
      return delta;
   } // method - Timer::getDelta
   
   void
   query_mount_points(  FrameQuery& query,
                        const char* ifo,
                        const char* type,
                        const unsigned int start_time,
                        const unsigned int stop_time )
   {
#if WORKING
      // Get current MOUNT_PT list
      const vector< string >
         mount_pt_list( MountPointHash::getMountPtList( ) );
   

      // Fix for problem report #2349:   
      // need to do a substring match for provided single character "ifo" - 
      // match any ifo that contains that single character
      if( strlen( ifo ) == 1 )
      {
         // Look up frame data under each MOUNT_PT entry
         for( vector< string >::const_iterator
                 iter = mount_pt_list.begin(),
                 end_iter = mount_pt_list.end();
              iter != end_iter;
              ++iter )
         {
            // FIX PR 2390: new query method: pass start,stop into search
            if( query( MountPointHash::
                       getIFOTypeHashes( *ifo,
                                         type,
                                         *iter,
                                         start_time, stop_time ) ) )
            {
               break;
            }
         }      
      }
      else
      {   
         const string ifo_type( query.getIFOType() );      
      
         // Look up frame data under each MOUNT_PT entry
         for( vector< string >::const_iterator
                 iter = mount_pt_list.begin(),
                 end_iter = mount_pt_list.end();
              iter != end_iter;
              ++iter )
         {
            // FIX PR 2390: new query method: pass start,stop into search
            if( query( MountPointHash::
                       getIFOTypeHashes( ifo_type,
                                         *iter,
                                         start_time, stop_time ) ) )
            {
               break;
            }
         }
      }
#endif /* WORKING */
   } // function - query_mount_points


   inline void
   is_first_scan_complete( const char* const File,
                           int Line )
   {
      if ( diskCache::MountPointScanner::FirstScanComplete( )
           == false )
      {
         std::ostringstream     msg;

         msg << "The in memory database of the file systems is currently unavailable"
             << " due to a request to do a full resync."
            ;

         std::ostringstream     info;
         info << "There are several reasons this may occur."
              << " 1. When the diskcacheAPI is restarted."
              << " 2. When a request has been made to rebuild the cache."
              << " 3. When the list of mount points has been changed."
              << " In all of these cases, the system must process each entry"
              << " in the list of mount points once before resuming normal"
              << " system processing."
            ;
         throw LdasException( Library::DISKCACHEAPI, -1,
                              msg.str( ), info.str( ),
                              File, Line );
      }
   }
      
   //====================================================================================
   //
   //====================================================================================
   std::string
   ms_format( const General::GPSTime& DT )
   {
      std::ostringstream	msg;

      INT_4U sec( DT.GetSeconds( ) );
      INT_4U hrs( sec / 3600 );
      sec %= 3600;
      INT_4U min( sec / 60 );
      sec %= 60;
      INT_4U ms = DT.GetNanoseconds( ) / 1000000;


      msg.fill( '0' );

      msg << std::setw( 2 ) << hrs
          << ":" << std::setw( 2 ) << min
          << ":" << std::setw( 2 ) << sec
          << "." << std::setw( 3 ) << ms
         ;

      return msg.str( );
   }

} // namespace - anonymous
   
   
//-----------------------------------------------------------------------
///
/// \brief Gets directory information that is listed in MOUNT_PT list.
///   
/// This method is used to introduce directory to the global hash map and to
/// rescan existing hash entry for any changes.
///   
/// \param[in] dirpath
///     Directory path to examine.
///
/// \return
/// A space separated list of all identified and scanned subdirectories.  
///
//-----------------------------------------------------------------------
string
getDirEnt( const std::string dirpath )
{
   Timer tracker;
   
   const string result = MountPointHash::addEntry( dirpath );         
   
   string msg( "getDirEnt( \"" );
   msg += dirpath;
   msg += "\" )";
   tracker.getDelta( msg );

   return result;
}
  
#if defined( CREATE_THREADED1 )
CREATE_THREADED1( getDirEnt, std::string, const std::string )
#endif /* defined( CREATE_THREADED1 ) */

   
   
//-----------------------------------------------------------------------
/// Update list of mount point directories (as they appear in
/// the resource file): initialize list if called at API startup time;
///                     otherwise, check if any MOUNT_PT directory
///                     has been pruned, and remove that directory from
///                     global hash.   
///
/// \param[in] dir_list
///       A list of directories as they appear
///       in MOUNT_PT list (API resource file variable).
/// \param[in] enable_global_check
///       A flag to indicate if API should check for data conflicts
///       across all MOUNT_PT entries (diskcacheAPI\@LDAS reports data
///       conflicts only under the same  MOUNT_PT).
///       This flag should be set to TRUE only when called by outside of 
///       LDAS utility, such as /ldas/bin/cacheCheck.   
///       Default is FALSE.   
///   
/// \return Tcl list of all removed subdirectories.
//-----------------------------------------------------------------------
string
updateMountPtList( const CHAR* dir_list,
                   const bool enable_global_check )
{
   Timer tracker;   
   
   const string
      result( MountPointHash::
              updateMountPtList( dir_list, enable_global_check ) );
   
   string msg( "updateMountPtList( \"" );
   msg += dir_list;
   msg += "\" )";
   tracker.getDelta( msg );

   return result;
}
   
   
//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
string
getDirCache( const char* ifo, const char* type )
{
   return getDirCache( true, ifo, type );
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
string
getDirCache( bool Sync, const char* ifo, const char* type )
{
   static const char caller[]= "getDirCache";

   if ( Sync )
   {
      is_first_scan_complete( __FILE__, __LINE__ );
   }

   Timer tracker;   
   
   string data;
   //--------------------------------------------------------------------
   // Set the search criteria
   //--------------------------------------------------------------------
   Query             q;
   QueryResultsDump  results( QueryResults::FORMAT_TCL );

   q.FNHandler( SDGTxFilenameHandler::Key( ) );
   q.IFO( ifo );
   q.Type( type );
   q.Extension( MountPointHash::getFileExtList( ).front( ) );

   {
      std::ostringstream        msg;

      msg << "Searching for ifo: " << ifo
          << " type: " << type
          << " extension: " << MountPointHash::getFileExtList( ).front( )
         ;
      GenericAPI::queueLogEntry( msg.str( ),
                                 GenericAPI::LogEntryGroup_type::MT_DEBUG,
                                 50,
                                 caller,
                                 "SEARCH_MOUNTPT" );
   }

   MountPointHash::Find( q, results );

   {
      std::ostringstream        msg;

      msg << "Operation took " << tracker.getDelta( caller )
          << " seconds to execute."
         ;
      GenericAPI::queueLogEntry( msg.str( ),
                                 GenericAPI::LogEntryGroup_type::MT_DEBUG,
                                 50,
                                 caller,
                                 "SEARCH_MOUNTPT" );
   }
   results.Finalize( );
   {
      std::ostringstream        msg;

      msg << "Finished the Finalization phase after executing for " << tracker.getDelta( caller )
          << " seconds."
         ;
      GenericAPI::queueLogEntry( msg.str( ),
                                 GenericAPI::LogEntryGroup_type::MT_DEBUG,
                                 50,
                                 caller,
                                 "SEARCH_MOUNTPT" );
   }
   results.Swap( data );
   {
      std::ostringstream        msg;

      msg << "Finished the Swapping phase after executing for " << tracker.getDelta( caller )
          << " seconds."
         ;
      GenericAPI::queueLogEntry( msg.str( ),
                                 GenericAPI::LogEntryGroup_type::MT_DEBUG,
                                 50,
                                 caller,
                                 "SEARCH_MOUNTPT" );
   }
   
   tracker.getDelta( "getDirCache()" );
   
   return data;
}
#if defined( CREATE_THREADED2 )
CREATE_THREADED2( getDirCache, std::string, const char*, const char* )
#endif /* defined( CREATE_THREADED2 ) */


//-----------------------------------------------------------------------
///
/// Get TCL formatted lists for each mount point with name or mount point,
/// number of directories and number of files for data matching the specified
/// ifo and type.  Return lists are of the form:
///   {mountpoint_name number_of_dirs number_of_files }
///
/// \param[in] ifo
///     An ifo to look up. Default is "all".
/// \param[in] type
///     A type to look up. Default is "all".
//
/// \return
///     A list for each mount point with name, number of
///     directories and number of files under that mount point:
///     {mountpoint_name number_of_dirs number_of files }
///
//-----------------------------------------------------------------------
string
getHashNumbers( const char* ifo, const char* type )
{
   is_first_scan_complete( __FILE__, __LINE__ );

   // Debug timer
   Timer tracker;

   string data;
   static const string all( "all" );
   
   data = MountPointHash::getHashNumbers( ifo, type );

   tracker.getDelta( "getHashNumbers()" );   

   return data;
}

//------------------------------------------------------------------------------
//
//: Get TCL formatted lists of file names for data matching the specified
//: IFO and Type within the bounds of the query.
//
//!usage: set string [ getFileNames ifo_type_str query_start query_stop ]
//
//!param: const char* ifo_type_str - A space delimited list of IFO-Type strings
//!param: const unsigned int start_time - query start time.
//!param: const unsigned int stop_time - query stop time.
//
//!return: string - A list for each mount point with data file names:
//+ { mntpt1 { /f1 /f2 ... /fN } mntpt2 { /f1 /f2 ... /fN } ... mntptN { } } 
//
string getFileNames( const char* ifo_type_str,
                     const unsigned int query_start,
                     const unsigned int query_stop)
{
   is_first_scan_complete( __FILE__, __LINE__ );

   Timer tracker;

   // string to fill with data and return
   string record("");

   record = MountPointHash::getFileNames( ifo_type_str, 
                                          query_start, 
                                          query_stop);

   tracker.getDelta( "getFileNames()" );

   return record;
}


//-----------------------------------------------------------------------
/// \brief Get TCL formatted lists of intervals for data matching
///        the specified and Type within the bounds of the query.
///
/// \param[in] ifo_type_str
///     A space delimited list of IFO-Type strings
/// \param[in] query_start
///     Query start time.
/// \param[in] query_stop
///     Query stop time.
///
/// \return
///     A list for each IFO-Type with data intervals:
///     IFO-Type1 { i1_start i1_stop ... iN_start iN_stop }
///     ... IFO-TypeN { ... }
//-----------------------------------------------------------------------
string
getIntervalsList( const char* ifo_type_str,
                  const unsigned int query_start,
                  const unsigned int query_stop )
{
   is_first_scan_complete( __FILE__, __LINE__ );

   Timer tracker;  

   // IFO Type delimiter
   static const char dash( '-' );

   // Copy and cast the passed in character array into a string object
   string raw_string(ifo_type_str);

   // string to fill with data and return
   string record("");
   
   // string buffer
   string buffer;

   // Insert the raw string into a stream to easily tokenize on whitespace
   istringstream ss(raw_string); 

   // Tokenize the string on whitespace
   while (ss >> buffer)
   {
      // Strings to fill from tokenizing the raw_string
      string ifo("");
      string type("");

      // locate dash in IFO-Type string
      string::size_type pos = buffer.find(dash, 0);

      // Extract IFO and Type component sub-strings
      ifo  = buffer.substr( 0, pos );
      type = buffer.substr( pos+1 );

      // get the interval data for this IFO-Type Combo
      record += " ";
      record += MountPointHash::
         getIntervalsList( ifo, type, query_start, query_stop);
   }

   tracker.getDelta( "getIntervalsList()" );

   return record;
}

#if FOR_LDAS
/// \cond EXCLUDE
CREATE_THREADED3( getIntervalsList, std::string, const char*,
		  const unsigned int, const unsigned int )
/// \endcond
#endif /* FOR_LDAS */


//-----------------------------------------------------------------------
/// \brief Frame query.
///   
/// This function is used to search global frame data hash for a
/// specific data.
///   
/// \param ifo
///     Data IFO.    
/// \param type
///     Data Type.       
/// \param start_time
///     Data start time.
/// \param stop_time
///     Data stop time.   
/// \param gaps_allowed
///     Flag to indicate if data gaps are allowed.
///     Set to true (1) if allowed, to false (0) if not.
///
/// \return
///     Tcl formated string: {frame files} {errors if any}.
//-----------------------------------------------------------------------
string
getFrameFiles( const char* ifo, const char* type,
               const unsigned int start_time,
               const unsigned int stop_time,
               const bool gaps_allowed )
{
   // Debug timer
   Timer tracker;   

   std::string                  filenames;

   try
   {
      is_first_scan_complete( __FILE__, __LINE__ );
   }
   catch( const LdasException& Exception )
   {
      const ErrorInfo& info( Exception[ 0 ] );

      std::ostringstream        msg;

      msg << "{} {" << info.getMessage( ) << "}";
      filenames = msg.str( );

      tracker.getDelta( "getFrameFiles()" );
      return filenames;
   }

   // User treats stop time inclusively, diskcache C++ exclusively, 
   // always increment user's stop time by 1 second:
   const unsigned int           actual_stop_time( stop_time + 1 );
   Query                        q;
   QueryResultsTimeRange        results( QueryResults::FORMAT_TCL, gaps_allowed );

   q.FNHandler( SDGTxFilenameHandler::Key( ) );
   q.IFO( ifo );
   q.Type( type );
   q.TimeRange( start_time, actual_stop_time );
   q.Extension( MountPointHash::getFileExtList( ).front( ) );

   MountPointHash::Find( q, results );

   results.Finalize( q );
   results.Swap( filenames );

   tracker.getDelta( "getFrameFiles()" );   
   return filenames;
}
   

#if FOR_LDAS
CREATE_THREADED5( getFrameFiles, std::string, const char*, const char*,
                  const unsigned int, const unsigned int, const bool )

#endif /* ! FOR_PYTHON && ! FOR_TCL */
   

//----------------------------------------------------------------------------
//   
//: RDS frame query.
//   
// This command finds frame files that satisfy createRDS user request. 
// If "will_resample" flag is set to TRUE, requested time range will be
// automatically expanded by frame dt at the beginning and at the end.
//   
//!param: const char* ifo - Data IFO.    
//!param: const char* type - Data Type.       
//!param: const unsigned int start_time - Data start time.
//!param: const unsigned int stop_time - Data stop time.   
//!param: const bool will_resample - Flag to indicate if requested data will be
//+                                  resampled. Set to TRUE (1) if will be 
//+                                  resampled, to FALSE (0) if not.
//!return: string - Tcl formated string: { frame files }{ errors if any }.
//                       
string
getRDSFrameFiles( const char* ifo,
                  const char* type,
                  const unsigned int start_time,
                  const unsigned int stop_time,
                  const bool will_resample )
{
   static const char caller[]= "getRDSFrmaeFiles";
   // Debug timer
   Timer tracker;   

   std::string                  filenames;

   try
   {
      is_first_scan_complete( __FILE__, __LINE__ );
   }
   catch( const LdasException& Exception )
   {
      const ErrorInfo& info( Exception[ 0 ] );

      std::ostringstream        msg;

      msg << "{} {" << info.getMessage( ) << "}";
      filenames = msg.str( );

      tracker.getDelta( "getFrameFiles()" );
      return filenames;
   }

   // User treats stop time inclusively, diskcache C++ exclusively, 
   // always increment user's stop time by 1 second:
   const unsigned int           actual_stop_time( stop_time + 1 );
   Query                        q;
   QueryResultsTimeRange        results( QueryResults::FORMAT_TCL, false );

   q.FNHandler( SDGTxFilenameHandler::Key( ) );
   q.IFO( ifo );
   q.Type( type );
   q.TimeRange( start_time, actual_stop_time );
   q.Extension( MountPointHash::getFileExtList( ).front( ) );

   std::ostringstream   msg;

   msg << "Search criteria: ifo: " << q.IFO( )
       << " type: " << q.Type( )
       << " time range: " << q.TimeRange( ).begin( )->first
       << "-" << q.TimeRange( ).begin( )->second
       << " extension: " << q.Extension( )
      ;
   GenericAPI::queueLogEntry( msg.str( ),
                              GenericAPI::LogEntryGroup_type::MT_DEBUG,
                              0,
                              caller,
                              "SEARCH_MOUNTPT" );
   MountPointHash::Find( q, results );

   if ( will_resample )
   {
      //-------------------------------------------------------------------------
      // Need to request an extra file on each side of the origional query
      // to allow for resampling
      //-------------------------------------------------------------------------
      if ( results.Resample( q ) )
      {
         MountPointHash::Find( q, results );
      }
   }

   results.Finalize( q );
   results.Swap( filenames );

   tracker.getDelta( "getFrameFiles()" );   

   msg.str( filenames );
   GenericAPI::queueLogEntry( msg.str( ),
                              GenericAPI::LogEntryGroup_type::MT_DEBUG,
                              0,
                              caller,
                              "SEARCH_MOUNTPT" );

   return filenames;
}
   

#if FOR_LDAS
CREATE_THREADED5( getRDSFrameFiles, std::string, const char*, const char*,
		  const unsigned int, const unsigned int, const bool )

#endif /* FOR_LDAS */


   
//----------------------------------------------------------------------------
//   
//: Write content of global frame data hash to binary file.   
//
//!param: const char* filename - Name of the file to write frame hash to.
//+       Default is NULL (C++ will use default file name).      
//
//!return: Nothing.
//
void
writeDirCache( const char* filename ) 
{
#if 0
   MountPointScanner::SyncRequest sync_request;
#endif /* 0 */
   MountPointScanner::scanner_sync_ro_type
      scanner_sync_lock( MountPointScanner::SyncRO( ) );

   // Debug timer
   Timer tracker;   
   
   CacheFile    cf;

   cf.writeToFile( filename );

   tracker.getDelta( "writeDirCache()" );   
   
   return;
}
   
   
#if defined( CREATE_THREADED1V )

CREATE_THREADED1V( writeDirCache, const char* )
   
#endif /* defined( CREATE_THREADED1V */
   
//----------------------------------------------------------------------------
//   
//: Write content of global frame data hash to ascii file.   
//
//!param: const char* filename - Name of the file to write frame hash to.
//+       Default is NULL (C++ will use default file name).      
//
//!return: Nothing.
//
   void writeDirCacheAscii( const char* filename ) 
{
#if 0
   MountPointScanner::SyncRequest sync_request;
#endif /* 0 */
   MountPointScanner::scanner_sync_ro_type
      scanner_sync_lock( MountPointScanner::SyncRO( ) );

   static const CHAR* method_name( "writeDirCacheAscii" );
   static const CHAR* const DefaultFileName( "frame_cache_dump2" );
   static const string TmpExtension( ".tmp" );
   static const string BackupExtension( ".bak" );   
   static MutexLock::lock_type key = MutexLock::Initialize( );

   // Debug timer
   Timer tracker;
   MutexLock    baton( key );
   std::string  fn( DefaultFileName );

   if ( filename && *filename )
   {
      fn = filename;
   }
   string fn_tmp( fn );
   fn_tmp += TmpExtension;
   
   CacheFile::stream_type cache_file;

   cache_file.open( fn_tmp.c_str(), ios_base::out | ios_base::trunc );

   try
   {
      std::string hash( getDirCache( ) );
      std::string::size_type end = 0;
      std::string::size_type start = 0;
      while ( end != std::string::npos )
      {
         end = hash.find( " /", start );
         const std::string::size_type len( ( end == std::string::npos )
                                           ? end
                                           : ( end - start ) );
         cache_file << hash.substr( start, len ) << std::endl;
         start = end + 1;
      }
      cache_file.close( );
   }
   catch( ... )
   {
      cache_file.clear( );              // Clear any errors
      cache_file.close( );              // Close the file
      unlink( fn_tmp.c_str( ) );        // Remove the file
      throw;                            // rethrow the exception
   }

   // Create backup copy of existing frame cache file
   cache_file.open( fn.c_str( ), ios_base::in );
   
   if( cache_file )
   {
      // There is existing cache file ---> need to save it
      cache_file.close();

   
      // Format name for backup file
      string backup_name( fn );
      backup_name += BackupExtension;

   
      // Check if there is a backup file from previous write:
      ifstream backup_file( backup_name.c_str() );
   
      if( backup_file )
      {
         backup_file.close();
   
         if( unlink( backup_name.c_str() ) )
         {
            // Could not remove old backup file
            const string errno_msg( strerror( errno ) );
      
            string msg( method_name );
            msg += "Could not remove old backup copy of cache file \'";
            msg += backup_name;
            msg += "\' (errno=";
            msg += errno_msg;
            msg += "). New cache file is stored as temporary \'";
            msg += fn_tmp;
            msg += '\'';
      
            throw runtime_error( msg.c_str() );   
         }
      }
   
   
      if( rename( fn.c_str(), backup_name.c_str() ) )
      {
         // Could not rename file
         const string errno_msg( strerror( errno ) );
   
         // Creating backup copy failed:
         string msg( method_name );
         msg += "Could not create backup copy of already existing cache file \'";
         msg += fn;
         msg += "\' (errno=";
         msg += errno_msg;
         msg += "). New cache file is stored as temporary \'";
         msg += fn_tmp;
         msg += '\'';
   
         throw runtime_error( msg.c_str() );
      }
   }

   
   // Move file to the requested destination
   if( rename( fn_tmp.c_str(), fn.c_str() ) )
   {
      // Could not rename file
      const string errno_msg( strerror( errno ) );
   
      // Creating backup copy failed:
      string msg( method_name );
      msg += "Could not move temporary copy of cache file \'";
      msg += fn_tmp;
      msg += "\' (errno=";
      msg += errno_msg;
      msg += ") to the destination \'";
      msg += fn;
      msg += '\'';
   
      throw runtime_error( msg.c_str() );   
   }

   tracker.getDelta( "writeDirCacheAscii()" );   
   
   return;
}
   
   
#if defined( CREATE_THREADED1V )

CREATE_THREADED1V( writeDirCacheAscii, const char* )
   
#endif /* define( CREATE_THREADED1V ) */
   
//-----------------------------------------------------------------------
///
/// \brief Write content of global frame data hash to ascii file.   
///
/// \param[in] bfilename
///     Name of the file to write the binary frame hash to.
///     Default is NULL (C++ will use default file name).      
///
/// \param[in] afilename
///     Name of the file to write the ascii frame hash to.
///     Default is NULL (C++ will use default file name).      
///
//-----------------------------------------------------------------------
   void writeDirCacheFiles( const char* bfilename,
                            const char* afilename ) 
{
   // Debug timer
   Timer        tracker;
#if FOR_TCL
   tid*         binary_tid  = writeDirCache_t( bfilename,
                                               (Tcl_Interp*)NULL,
                                               "" );
   tid*         ascii_tid = writeDirCacheAscii_t( afilename,
                                                  (Tcl_Interp*)NULL,
                                                  "" );
   
   binary_tid->Join( ); // Syncronize point
   ascii_tid->Join( );  // Syncronize point
   
   writeDirCache_r( binary_tid );       // Cleanup
   writeDirCacheAscii_r( ascii_tid );   // Cleanup
#else
   writeDirCache( bfilename );
   writeDirCacheAscii( afilename );
#endif


   tracker.getDelta( "writeDirCacheFiles()" );   
   
   return;
}
   
   
#if defined( CREATE_THREADED2V )

CREATE_THREADED2V( writeDirCacheFiles, const char*, const char* )
   
#endif /* defined( CREATE_THREADED2V ) */
   
//------------------------------------------------------------------------------
//   
//: Read content of global frame data hash from binary file.   
//
// Tcl layer can specify different files for read and write operations. 
// This function forces all read and write operations to be sequencial.
//
// ATTN: This function destroys existing hash before reading a new one from the
//       given file. Caller must assure there are no running threads that might
//       access global frame data hash at the time "readDirCache" is called.   
//    
//!param: const char* filename - Name of the file to read frame hash from. 
//+       Default is NULL (C++ will use default file name).   
//
//!return: Nothing.
//   
void readDirCache( const char* filename )
{
   static const char caller[]= "readDirCache";

   // Debug timer
   Timer tracker;   

   {
      std::ostringstream        msg;

      msg << "Reading " << filename
         ;
      GenericAPI::queueLogEntry( msg.str( ),
                                 GenericAPI::LogEntryGroup_type::MT_DEBUG,
                                 50,
                                 caller,
                                 "SEARCH_MOUNTPT" );
   }

   //--------------------------------------------------------------------
   // Preserve the old value of the mount point list
   //--------------------------------------------------------------------
   MountPointHash::mount_point_container_type
      old_mount_points( MountPointHash::MountPoints( ).Var( ) );

   {
      //-----------------------------------------------------------------
      // Enscure serializatioin of reads and writes
      //-----------------------------------------------------------------
      DumpCacheDaemon::io_lock_type 
         lock( DumpCacheDaemon::IOLock( ) );

      CacheFile    cf;

      cf.readFromFile( filename );

      {
         std::ostringstream        msg;

         msg << "Read from file took " << tracker.getDelta( caller )
             << " seconds to execute."
            ;
         GenericAPI::queueLogEntry( msg.str( ),
                                    GenericAPI::LogEntryGroup_type::MT_DEBUG,
                                    50,
                                    caller,
                                    "SEARCH_MOUNTPT" );
      }
   }
   //--------------------------------------------------------------------
   // Restore the old value of the mount point list
   //--------------------------------------------------------------------
   {
      mount_point_container_type added;
      mount_point_container_type removed;
      std::list< mount_point_container_type::value_type >
         removed_subdirs;
      LogMessage errors;

      MountPointHash::UpdateMountPointList( old_mount_points,
                                            added,
                                            removed,
                                            removed_subdirs,
                                            errors );
   }

   tracker.getDelta( "readDirCache()" );   
   
   
   return;
}
   
#if defined( CREATE_THREADED1V )

CREATE_THREADED1V( readDirCache, const char* )
   
#endif /* defined( CREATE_THREADED1V ) */
   
//----------------------------------------------------------------------------
//   
//: Update list of excluded directories (as they appear in
//: the resource file): check if existing data hash relies on such directories
//:                     already, remove those directories recursively from
//:                     global hash.   
//
//!usage: set dir_list [ excludedDirList dirs_to_exclude ]
//   
//!param: const CHAR* dir_list - A list of directories to exclude (as they 
//+                              appear in API resource file variable).
//   
//!return: string - Sorted Tcl list of all removed subdirectories, followed
//+        by error messages if any:
//+        {Directories excluded by pattern 'dir_list': dir1 dir2 ...} 
//+        {error1 error2 ...}
//         
   string excludedDirList( const char* dir_list )
{
   // Debug timer
   Timer tracker;   
   
   const string result( MountPointHash::updateExcludedDirList( dir_list ) );

   tracker.getDelta( "excludedDirList()" );   
   
   
   return result;
}


//----------------------------------------------------------------------------
//
//: Update list of file name extensions 
//
//!usage: updateFileExtList [ list ]
//  
//!param: const CHAR* ext_list - A list of file extensions
//  
//!return: nothing 
//
void updateFileExtList( const char* ext_list )
{
   // Debug timer
   Timer tracker;

   MountPointHash::updateFileExtList( ext_list );

   tracker.getDelta( "updateFileExtList()" );

   return;
}


//----------------------------------------------------------------------------
//
//: Obtain the list of file name extensions
//
//!usage: set ext_list [ getFileExtList ]
// 
//!param: none 
// 
//!return: std::string
//
std::string getFileExtList()
{
   const std::list< std::string > file_ext_list(
                                                MountPointHash::getFileExtList() );

   ostringstream out_stream;

   for( std::list< std::string >::const_iterator iter = file_ext_list.begin(),
           end_iter = file_ext_list.end(); 
        iter != end_iter; 
        ++iter )
   {
      out_stream << (*iter);
      out_stream << " ";
   }

   return ( out_stream.str() );
}
   
//------------------------------------------------------------------------------
//   
//: Delete global frame data hash.
//
// ATTN: This function destroys existing hash in memory!
//       Caller must assure there are no running threads that might
//       access global frame data hash at the time "deleteDirCache" is called.
//    
//!return: Nothing.
//
void deleteDirCache()
{
   // Debug timer
   Timer tracker;   
   
   MountPointHash::removeEntries();

   tracker.getDelta( "deleteDirCache()" );   
   
   
   return;   
}
   
#if defined( CREATE_THREADED0V )
CREATE_THREADED0V( deleteDirCache )
#endif /* defined( CREATE_THREADED0V ) */

#if FOR_TCL
//-----------------------------------------------------------------------
/// \brief Bind a Tcl Varialbe to the C++ variable for mount pt conflict checking
///
/// \param Interp
///     The TCL interpreter containing the Variable.
/// \param Variable
///     name of variable to be used within TCL to control if
///	new time hashes need to be verified agaist all existing
///	time hashes.
///   
/// \return 1 if binding took place, 0 otherwise.
///
/// \todo Remove this function as the back end has been removed.
bool
bindToCheckForMountPTConflictVariable( Tcl_Interp* Interp, char* Variable )
{
   bool retval = false;
   return retval;
}
#endif /* FOR_TCL */

//-----------------------------------------------------------------------
/// This routine does a single pass of the listed mount points.
//-----------------------------------------------------------------------
void
ScanConcurrency( INT_4U Concurrency )
{
   MountPointScanner::Concurrency( Concurrency );
}

//-----------------------------------------------------------------------
/// This routine does a single pass of the listed mount points.
//-----------------------------------------------------------------------
void
ScanMountPointList( std::ostream* Stream )
{
   static const char caller[] = "ScanMountPointList";
   //--------------------------------------------------------------------
   /// Obtain the list of mount points.
   //--------------------------------------------------------------------
   General::GPSTime     start;
   start.Now( );

   diskCache::MountPointHash::
      ScannedMountPoints( diskCache::MountPointHash::SCANNED_MOUNT_POINTS_RESET );
   diskCache::DirectoryHash::
      ScannedDirectories( diskCache::DirectoryHash::SCANNED_DIRECTORIES_RESET );
   diskCache::ScannedFiles( diskCache::SCANNED_FILES_RESET );

   //--------------------------------------------------------------------
   // Loop over the mount points
   //--------------------------------------------------------------------
   MountPointScanner::Scan( diskCache::MountPointHash::MountPoints( ).Var( ), Stream );

   General::GPSTime     end;
   end.Now( );
   std::ostringstream   msg;
   General::GPSTime     dt( General::GPSTime( ) + ( end - start ) );
   INT_4U ms = dt.GetNanoseconds( ) / 1000000;
   ms += dt.GetSeconds( ) * 1000;
   //--------------------------------------------------------------------
   /// The output needs to be in the following format
   /// 7 mount points, 0 directories, 0 files, scanned in 3010 ms  (00:00:03.0010).
   //--------------------------------------------------------------------

   msg << diskCache::MountPointHash::ScannedMountPoints( ) << " mount points, "
       << diskCache::DirectoryHash::ScannedDirectories( ) << " directories, "
       << diskCache::ScannedFiles( ) << " files,"
       << " scanned in " << ms << " ms "
       << "(" << ms_format( dt ) << ")."
      ;
   if ( General::AtExit::IsExiting( ) == false )
   {
      if ( Stream )
      {
         (*Stream) << msg.str( ) << std::endl;
      }
      else
      {
         GenericAPI::queueLogEntry( msg.str( ),
                                    GenericAPI::LogEntryGroup_type::MT_NOTE,
                                    0,
                                    caller,
                                    "SCAN_MOUNTPT" );
      }
   }

}

//-----------------------------------------------------------------------
/// Create an image of the cache file on the storage media.
//-----------------------------------------------------------------------
void
DumpCacheDaemonStart( std::ostream* Stream )
{
   using namespace diskCache;

   //--------------------------------------------------------------------
   /// \todo
   /// Protect from multiple scanner daemons from running.
   //--------------------------------------------------------------------
   //--------------------------------------------------------------------
   // Get a thread in which to run this task in the background
   //--------------------------------------------------------------------
   General::TaskThread*         thread = General::ThreadPool::Acquire( );
   DumpCacheDaemon*             task = new DumpCacheDaemon( Stream );
   //--------------------------------------------------------------------
   /// \todo Need to setup the task thread to be returned to the pool
   ///     once the task is complete.
   //--------------------------------------------------------------------
   //--------------------------------------------------------------------
   // Add the task to the thread and get things running
   //--------------------------------------------------------------------
   thread->AddTask( task );
}

//-----------------------------------------------------------------------
/// Scan the list of mount points in the background
//-----------------------------------------------------------------------
void
ScanMountPointListContinuously( std::ostream* Stream )
{
   //--------------------------------------------------------------------
   /// \todo
   /// Protect from multiple scanner daemons from running.
   //--------------------------------------------------------------------
   //--------------------------------------------------------------------
   // Get a thread in which to run this task in the background
   //--------------------------------------------------------------------
   General::TaskThread*         thread = General::ThreadPool::Acquire( );
   ScanMountPointsDaemon*       task = new ScanMountPointsDaemon( Stream );
   //--------------------------------------------------------------------
   /// \todo Need to setup the task thread to be returned to the pool
   ///     once the task is complete.
   //--------------------------------------------------------------------
   //--------------------------------------------------------------------
   // Add the task to the thread and get things running
   //--------------------------------------------------------------------
   thread->AddTask( task );
}

// END 
