// Local Header Files
#include "filename.hh"
#include "mountpt.hh"

// General Header Files   
#include <general/autoarray.hh>   
   
// System Header Files   
#include <algorithm>
#include <sstream>   
   
using namespace std;
using namespace diskCache;
using namespace General;   
   

namespace 
{

} // local namespace
   
   
//------------------------------------------------------------------------------
//
//: Constructor.
//      
//!param: const std::string& file_name - File name to tokenize.
//!param: const std::string dir_path - A full path for the directory frame file
//+       resides in (includes terminating '/'; used only by exceptions).
//   
diskCache::FileName::FileName( const std::string& file_name,
   const std::string& dir_path )   
   : mName( file_name ), mDir( dir_path ), mStartTime( 0 ), mDt( 0 )
{}

   
//------------------------------------------------------------------------------
//
//: Should ignore the file?
//      
// This method determines if file name should be ignored while
// examining a directory it belongs to: 
// it starts with a '.' | temporary frame file.
//   
//!return: const bool - True if file name should be ignored,
//+        false otherwise.   
//   
const bool diskCache::FileName::shouldIgnore() const
{
   const std::list< std::string > file_ext_list(
        MountPointHash::getFileExtList() );

   const string file_extension( (file_ext_list.front() + ".tmp") );

   if( ( mName.size() && mName[ 0 ] == '.' ) ||
       ( hasExtension( file_extension, file_extension.size() ) ) )
   {
      return true;
   }
   
   return false;
}
   
   
//------------------------------------------------------------------------------
//
//: Is it a frame?
//      
// This method determines if file name has frame extension. If will extract
// necessary bits of information if it's a frame file.
//   
//!return: const bool - True if file is a frame, false otherwise.   
//   
const bool diskCache::FileName::isFrame()
{
   const std::list< std::string > file_ext_list(
        MountPointHash::getFileExtList() );

   const string file_extension( file_ext_list.front() );

   if( hasExtension( file_extension, file_extension.size() ) )
   {
      // Extract bits of information from file name
      parse();
   
      return true;
   }

   return false;
}

   
//------------------------------------------------------------------------------
//
//: Should exclude the file?
//      
// This method determines if file name should be excluded from the scan.
//
//!param: const std::list< std::string >& exclude_dirs - A list of known 
//+       file names to exclude from scan.
//   
//!return: const bool - True if file name should be excluded,
//+        false otherwise.   
//      
const bool diskCache::FileName::shouldExclude( 
   const excluded_dirs_container_type& exclude_dirs ) const   
{
   return ( find( exclude_dirs.begin(), exclude_dirs.end(), mName )
            != exclude_dirs.end() );
}
   
   
//------------------------------------------------------------------------------
//
//: Get "IFO-Type-" part of the file name.
//
//!return: const std::string - "IFO-Type-" part of the name.
//   
const std::string diskCache::FileName::getIFOType() const      
{
   return mIFOType;
}

   
//------------------------------------------------------------------------------
//
//: Get extracted GPS time at the beginning of the first frame in the file.
//
//!return: const INT_4U - Extracted GPS time.
//   
const INT_4U diskCache::FileName::getStartTime() const      
{
   return mStartTime;
}

   
//------------------------------------------------------------------------------
//
//: Get extracted total time interval covered by the file, in seconds.
//
//!return: const INT_4U - Extracted interval time.
//   
const INT_4U diskCache::FileName::getDt() const      
{
   return mDt;
}


//------------------------------------------------------------------------------
//   
//: Get frame name separator.   
//
//!return: const CHAR& - A reference to the character used as a frame name 
//+        separator.
//   
// static   
const CHAR& diskCache::FileName::getFrameNameSeparator()
{
   static const CHAR frame_separator( '-' );   
   return frame_separator;
}
   
   
//------------------------------------------------------------------------------
//
//: Does file have a specific extension?
//   
// This method checks if file name has a specific extension.
//   
//!param: const string& extension - Searched extension.
//!param: const size_t& extension_len - Length of searched extension.   
//
//!return: const bool - True if file has specified extension, false otherwise.   
//      
const bool diskCache::FileName::hasExtension( const std::string& extension,
   const size_t& extension_len ) const
{
   string::size_type ext_pos( mName.rfind( extension ) );

   if (ext_pos == string::npos) 
   {  
       return false; 
   }

   size_t file_len( mName.size() );
   
   if( file_len >= extension_len )
   {
      // Expected extension position for the current file
      file_len -= extension_len;
   }
   
   
   return( ext_pos == file_len );   
}
   
   
//------------------------------------------------------------------------------
//
//: Parse file name.
//      
// This method splits frame file name into required bits of information:
// "IFO-Type-", startTime, dT.   
//   
//!exc: runtime_error - Could not extract IFO information from file.
//!exc: runtime_error - Could not extract Type information from file.
//!exc: runtime_error - Error extracting StartTime information from file.
//!exc: runtime_error - Error extracting dT information from file.
//   
//!return: Nothing.
//   
void diskCache::FileName::parse()
{
   // diskcache-Tcl says ignore all frame files with malformed names,
   // capture as an error for now.
   static const CHAR* const method_name( "FileName::parse(): " );
   static const CHAR dash( '-' );
   static const CHAR dot( '.' );

   
   string::size_type pos( 0 );

   // Find IFO
   if( ( pos = mName.find( dash, pos ) ) == string::npos )
   {
      string msg( method_name );
      msg += "Could not extract IFO information from ";
      msg += mDir;
      msg += '/';
      msg += mName;
   
      throw runtime_error( msg.c_str() );
   }
   
   
   // Find Type
   if( ( pos = mName.find( dash, ++pos ) ) == string::npos )
   {
      string msg( method_name );
      msg += "Could not extract Type information from ";
      msg += mDir;
      msg += '/';   
      msg += mName;
   
      throw runtime_error( msg.c_str() );
   }   


   // Include 2nd dash: "IFO-Type-"
   mIFOType = mName.substr( 0, ++pos );
   

   // Create a stream for numerical extractions
   istringstream stream( mName.substr( pos ) );   
   
   // Extract GPS start time
   stream >> mStartTime;
   
   if( !stream || stream.peek() != dash )
   {
      string msg( method_name );
      msg += "Error extracting StartTime information from ";
      msg += mDir;
      msg += '/';
      msg += mName;
   
      throw runtime_error( msg.c_str() );   
   }
   
   // Skip delimiter
   stream.ignore();   

   
   // Extract time interval
   stream >> mDt;
   
   if( !stream || stream.peek() != dot )
   {
      string msg( method_name );
      msg += "Error extracting dT information from ";
      msg += mDir;
      msg += '/';
      msg += mName;
   
      throw runtime_error( msg.c_str() );   
   }

   
   return;
}
