#include "framecpp/config.h"

#include <cstdlib>

#include "framecpp/FrDetector.hh"
#include "framecpp/FrTOC.hh"

#include "framec/FrDetector.h"
#include "framec/Stream.h"

#include "FrameCInternal.hh"
#include "StreamInternal.hh"

using FrameC::Deposit;
using FrameC::Set;
using FrameC::Validate;
using FrameC::Withdraw;

namespace FrameC
{
  struct FrDetector {
    ::fr_detector_t			m_data;

    FrDetector( );

    FrDetector( const char* restrict name,
		const char* restrict prefix,
		fr_detector_latitude_t latitude,
		fr_detector_longitude_t longitude,
		fr_detector_elevation_t elevation,
		fr_detector_azimuth_x_t azimuth_x,
		fr_detector_azimuth_y_t azimuth_y,
		fr_detector_altitude_x_t altitude_x,
		fr_detector_altitude_y_t altitude_y,
		fr_detector_midpoint_x_t midpoint_x,
		fr_detector_midpoint_y_t midpoint_y,
		fr_detector_localtime_t localtime );

    ~FrDetector( );

    void Read( fr_file_t* restrict Stream,
	       const char* restrict Name );
  };


  void FrDetector::
  Read( fr_file_t* restrict Stream,
	const char* restrict Name )
  {
    FrameCPP::IFrameStream*	fs( StreamAsInput( Stream ) );

    std::auto_ptr< FrameCPP::Common::FrameSpec::Object >
      raw_detector( fs->ReadDetector( Name ) );

    std::auto_ptr< FrameCPP::FrDetector >
      d( dynamic_cast< FrameCPP::FrDetector* >( raw_detector.get( ) ) );

    if( d.get( ) )
    {
      //-----------------------------------------------------------------
      // Ensure only one smart pointer has control
      //-----------------------------------------------------------------
      raw_detector.release( );
      //-----------------------------------------------------------------
      // Move data into the C structure.
      //-----------------------------------------------------------------
      m_data.name = ::strdup( d->GetName( ).c_str( ) );
      ::strncpy( m_data.prefix, d->GetPrefix( ), DETECTOR_PREFIX_LEN );
      m_data.prefix[ DETECTOR_PREFIX_LEN ] = '\0';
      m_data.latitude = d->GetLatitude( );
      m_data.longitude = d->GetLongitude( );
      m_data.elevation = d->GetElevation( );
      m_data.azimuth_x = d->GetArmXazimuth( );
      m_data.azimuth_y = d->GetArmYazimuth( );
      m_data.altitude_x = d->GetArmXaltitude( );
      m_data.altitude_y = d->GetArmYaltitude( );
      m_data.midpoint_x = d->GetArmXmidpoint( );
      m_data.midpoint_y = d->GetArmYmidpoint( );
      m_data.localtime = d->GetLocalTime( );
    }
    else
    {
      std::ostringstream	msg;

      msg << "Unable to locate the detector named " << Name
	  << " within the stream.";

      throw std::range_error( msg.str( ) );
    }

  }

  template<>
  inline Pointers::pointer_type
  PointerType( FrDetector* Value )
  {
    return Pointers::FR_DETECTOR;
  }

  template<>
  const char* Pointers::
  Name< Pointers::FR_DETECTOR >( )
  {
    return "FrDetector";
  }
}

fr_detector_t*
FrDetectorAlloc( FrameCError** Error,
		 const char* restrict name,
		 const char* restrict prefix,
		 fr_detector_latitude_t latitude,
		 fr_detector_longitude_t longitude,
		 fr_detector_elevation_t elevation,
		 fr_detector_azimuth_x_t azimuth_x,
		 fr_detector_azimuth_y_t azimuth_y,
		 fr_detector_altitude_x_t altitude_x,
		 fr_detector_altitude_y_t altitude_y,
		 fr_detector_midpoint_x_t midpoint_x,
		 fr_detector_midpoint_y_t midpoint_y,
		 fr_detector_localtime_t localtime)
{
  std::auto_ptr< FrameC::FrDetector >	retval_sptr;

  try
  {
    retval_sptr.reset( new FrameC::FrDetector( name,
					       prefix,
					       latitude,
					       longitude,
					       elevation,
					       azimuth_x,
					       azimuth_y,
					       altitude_x,
					       altitude_y,
					       midpoint_x,
					       midpoint_y,
					       localtime ) );
  }
  catch( const std::exception& Exception )
  {
    Set( Error, FRAMEC_ERRNO_FR_DETECTOR_ALLOC_ERROR, Exception.what( ) );
    retval_sptr.reset( NULL );
  }
  catch( ... )
  {
    Set( Error, FRAMEC_ERRNO_FR_DETECTOR_ALLOC_ERROR, FrameC::UnknownError );
    retval_sptr.reset( NULL );
  }
  //---------------------------------------------------------------------
  // Report the results
  //---------------------------------------------------------------------
  fr_detector_t*	retval( (fr_detector_t*)NULL );

  if ( retval_sptr.get( ) )
  {
    Deposit( retval_sptr.get( ) );
    retval = &(retval_sptr->m_data );
    retval_sptr.release( );
    
  }
  return retval;
}

void
FrDetectorFree( FrameCError** Error,
		fr_detector_t* detector )
{
  Set( Error );

  FrameC::FrDetector*      oodetector( (FrameC::FrDetector*)detector );

  try
  {
    if ( oodetector )
    {
      Withdraw( oodetector );
    }
  }
  catch( const std::exception& Exception )
  {
    Set( Error, FRAMEC_ERRNO_FR_DETECTOR_FREE_ERROR, Exception.what( ) );
  }
}


fr_detector_t*
FrDetectorRead( FrameCError** Error,
		fr_file_t* restrict stream,
		const char* restrict name )
{
  Set( Error );

  std::auto_ptr< FrameC::FrDetector >	retval_sptr;

  try
  {
    retval_sptr.reset( new FrameC::FrDetector );
    retval_sptr->Read( stream, name );
  }
  catch( const std::exception& Exception )
  {
    Set( Error, FRAMEC_ERRNO_FR_DETECTOR_READ_ERROR, Exception.what( ) );
    retval_sptr.reset( NULL );
  }

  fr_detector_t*	retval = (fr_detector_t*)NULL;

  if ( retval_sptr.get( ) )
  {
    Deposit( retval_sptr.get( ) );
    retval = &(retval_sptr->m_data);
    retval_sptr.release( );
  }
  return retval;
}

int
FrDetectorWrite( FrameCError** Error,
		 fr_file_t* restrict stream,
		 const fr_detector_t* restrict detector )
{
  int retval = 1;

  return retval;
}

namespace FrameC
{
  FrDetector::
  FrDetector( )
  {
    m_data.name = (char*)NULL;
    m_data.prefix[ 0 ] = '\0';
  }

  FrDetector::
  FrDetector( const char* restrict name,
	      const char* restrict prefix,
	      fr_detector_latitude_t latitude,
	      fr_detector_longitude_t longitude,
	      fr_detector_elevation_t elevation,
	      fr_detector_azimuth_x_t azimuth_x,
	      fr_detector_azimuth_y_t azimuth_y,
	      fr_detector_altitude_x_t altitude_x,
	      fr_detector_altitude_y_t altitude_y,
	      fr_detector_midpoint_x_t midpoint_x,
	      fr_detector_midpoint_y_t midpoint_y,
	      fr_detector_localtime_t localtime )
  {
    m_data.name = ::strdup( name );
    ::strncpy( m_data.prefix, prefix, DETECTOR_PREFIX_LEN );
    m_data.prefix[ DETECTOR_PREFIX_LEN ] = '\0';
    m_data.latitude = latitude;
    m_data.longitude = longitude;
    m_data.elevation = elevation;
    m_data.azimuth_x = azimuth_x;
    m_data.azimuth_y = azimuth_y;
    m_data.altitude_x = altitude_x;
    m_data.altitude_y = altitude_y;
    m_data.midpoint_x = midpoint_x;
    m_data.midpoint_y = midpoint_y;
    m_data.localtime = localtime;
  }

  FrDetector::
  ~FrDetector( )
  {
    if ( m_data.name )
    {
      free ( m_data.name );
      m_data.name = (char*)NULL;
    }
  }
}
