#include "framecpp/config.h"

#include <cstdlib>

#include "framecpp/Common/FrTOC.hh"

#include "framec/FrTOC.h"
#include "framec/Stream.h"

#include "FrameCInternal.hh"
#include "StreamInternal.hh"

using FrameC::Deposit;
using FrameC::Set;
using FrameC::Validate;
using FrameC::Withdraw;

namespace FrameC
{
  struct FrTOC {
    ::fr_toc_t			m_data;

    FrTOC( );

    ~FrTOC( );
  };

  template<>
  inline Pointers::pointer_type
  PointerType( FrTOC* Value )
  {
    return Pointers::FR_TOC;
  }

  template<>
  const char* Pointers::
  Name< Pointers::FR_TOC >( )
  {
    return "FrTOC";
  }
}

namespace
{
  inline void
  copy_toc( const FrameCPP::Common::FrTOC* TOC,
	    FrameC::FrTOC* CTOC )
  {
    using FrameCPP::Common::FrTOC;

    const FrTOC::cmn_dt_container_type& 	dt( TOC->dt( ) );
    const FrTOC::cmn_GTimeS_container_type& 	gpss( TOC->GTimeS( ) );
    const FrTOC::cmn_GTimeN_container_type& 	gpsn( TOC->GTimeN( ) );

    const size_t		n( dt.size( ) );
    CTOC->m_data.nframe = n;

    CTOC->m_data.t0 = (fr_toc_t0_t*)calloc( n, sizeof( fr_toc_t0_t ) );
    CTOC->m_data.dt = (fr_toc_dt_t*)calloc( n, sizeof( fr_toc_dt_t ) );

    for ( size_t c = 0;
	  c != n;
	  ++c )
    {
      CTOC->m_data.t0[ c ].sec = gpss[ n ];
      CTOC->m_data.t0[ c ].nan = gpsn[ n ];
      CTOC->m_data.dt[ c ] = dt[ n ];
    }
    
  }
}

/**
 *  Read a subset of the table of contents information.
 *
 * \note
 *  Currently, this function is currently limited to input streams.
 */
fr_toc_t*
FrTOCRead( FrameCError** Error,
	   fr_file_t* Stream )
{
  std::auto_ptr< FrameC::FrTOC >	retval;

  try
  {
    retval.reset( new FrameC::FrTOC );


    retval->m_data.t0 = ( fr_toc_t0_t*)NULL;
    retval->m_data.dt = ( fr_toc_dt_t*)NULL;

    Validate( Stream );
    if ( Stream->s_mode == FRAMEC_FILE_MODE_INPUT )
    {
      //-----------------------------------------------------------------
      // input stream
      //-----------------------------------------------------------------

      FrameCPP::IFrameStream*	fs( Stream->s_stream.input );

      copy_toc( fs->GetTOC( ),
		retval.get( ) );
    }
    else
    {
#if WORKING
      //-----------------------------------------------------------------
      // output stream
      //-----------------------------------------------------------------
      FrameCPP::OFrameStream*	fs( Stream->s_stream.output );

      copy_toc( fs->GetTOC( ),
		retval.get( ) );
#else	/* WORKING */
      throw std::runtime_error( "Unable to obtain a table of contents for a frame being written" );
#endif	/* WORKING */
    }
  }
  catch( const std::exception& Exception )
  {
    Set( Error, FRAMEC_ERRNO_FR_TOC_READ_ERROR, Exception.what( ) );
    retval.reset( NULL );
  }
  catch( ... )
  {
    Set( Error, FRAMEC_ERRNO_FR_TOC_READ_ERROR, FrameC::UnknownError );
    retval.reset( NULL );
  }
  fr_toc_t* toc = (fr_toc_t*)NULL;
  if ( retval.get( ) )
  {
    Deposit( retval.get( ) );
    toc = &(retval->m_data );
    retval.release( );
    
  }
  return toc;
}

/**
 * This function simply recycles resources allocated by FrTOCRead.
 * Once the resource have been recyled, the resource are no longer
 * available to the application for use.
 */
void
FrTOCFree( FrameCError** Error,
	   fr_toc_t* TOC )
{
  Set( Error );

  FrameC::FrTOC*      ootoc( (FrameC::FrTOC*)TOC );

  try
  {
    if ( ootoc )
    {
      Withdraw( ootoc );
    }
  }
  catch( const std::exception& Exception )
  {
    Set( Error, FRAMEC_ERRNO_FR_TOC_FREE_ERROR, Exception.what( ) );
  }
}

namespace FrameC
{
  FrTOC::
  FrTOC( )
  {
  }

  FrTOC::
  ~FrTOC( )
  {
    if ( m_data.t0 )
    {
      free( m_data.t0 );
    }
    if ( m_data.dt )
    {
      free( m_data.dt );
    }
  }
}
